function score = evalSvMap(segPath, dbName, gtPath, gtScale, framePath, fmt, savepath, drawpath)
% This function is used to evaluate a supervoxel segmentation video.

%% Load Ground-Truth
fprintf('Loading Ground-Truth ... \n');
if ischar(gtScale)
    gtScale = str2double(gtScale);
end
switch dbName
    case 'BuffaloXiph'
        [gtMap, gtList, gtPos] = loadGT_BuffaloXiph(gtPath, gtScale);
        [gt_h, gt_w, ~] = size(gtMap);
    case 'SegTrackv2'
        [gtMap, gtList, gtPos] = loadGT_SegTrackv2(gtPath, gtScale);
        [gt_h, gt_w, ~] = size(gtMap);
    case 'CamVid'
        [gtMap, gtList, gtPos] = loadGT_CamVid(gtPath, gtScale);
        [gt_h, gt_w, ~] = size(gtMap);
    case 'BVDS'
        [gtMap, gtList, gtPos] = loadGT_BVDS(gtPath, gtScale);
        [gt_h, gt_w, ~] = size(gtMap{1});
    otherwise
        fprintf('Failed to load ground-truth. \n');
        return
end

%% Load Supervoxel Segmentation
fprintf('Loading Supervoxel Segmentation ... \n');
load(segPath); % svMap
svNum = double(max(svMap(:)));
[sv_h, sv_w, sv_f] = size(svMap);
if gt_h ~= sv_h || gt_w ~= sv_w
    tmpMap = [];
    for i=1:size(svMap,3)
        tmpMap = cat(3, tmpMap, imresize(svMap(:,:,i), [gt_h, gt_w], 'nearest'));
    end
    svMap = tmpMap;
end

% if svNum < 5
%     score = [];
%     fprintf('Warning: Too Few Supervoxels.\n');
%     return
% end

%% Average Superpixel Per Frame
spNumList = zeros(sv_f,1);
for i=1:sv_f
    spNumList(i) = length(unique(svMap(:,:,i)));
end
spNum = mean(spNumList);

% if spNum > 5000
%     score = [];
%     fprintf('Warning: Too Many Supervoxels.\n');
%     return
% end


%% Supervoxel Evaluation Metrics
fprintf('Evaluating .. \n')
[TEX, SZV] = scoreMeanDurationAndSizeVariation(svMap);
EV = scoreExplainedVariation(svMap, framePath, fmt);

if strcmp(dbName, 'BVDS')
    % multiple human annotations
    gtPeopleNum = length(gtMap);
    mUE3D = zeros(gtPeopleNum,1); mSA3D = zeros(gtPeopleNum,1);
    mBRD = zeros(gtPeopleNum,1);
    for i=1:gtPeopleNum
        [mUE3D(i), mSA3D(i)] = scoreUnderErrorAndSegAccuracy(svMap, gtMap{i}, gtList{i}, gtPos, '3D');
        thisBRD = scoreBoundaryRecallDistance(svMap, gtMap{i}, gtPos);
        if isempty(thisBRD)
            score = [];
            fprintf('Warning: Bad supervoxel in %s!\n', segPath);
            return
        else
            mBRD(i) = thisBRD;
        end

    end
    UE3D = mean(mUE3D); SA3D = mean(mSA3D);
    BRD = mean(mBRD);
else
    % single human annotation
    [UE3D, SA3D, onMapUE, onMapSA] = scoreUnderErrorAndSegAccuracy(svMap, gtMap, gtList, gtPos, '3D');
    BRD = scoreBoundaryRecallDistance(svMap, gtMap, gtPos);
%     [BRD, gtBrWithin] = scoreBoundaryRecallDistance(svMap, gtMap, gtPos);
    if isempty(BRD)
        score = [];
        fprintf('Warning: Bad supervoxel in %s!\n', segPath);
        return
    end
end

%% Output Score
fprintf('|%6s|%6s|%6s|%6s|%6s|%6s|%6s|%6s|\n', ...
    'svNum', 'spNum', 'TEX', 'SZV', 'EV', 'UE3D', 'SA3D', 'BRD');
fprintf('|%6.2f|%6.2f|%6.2f|%6.2f|%6.2f|%6.2f|%6.2f|%6.2f|\n', ...
    svNum, spNum, TEX, SZV, EV, UE3D, SA3D, BRD);
score = [svNum; spNum; TEX; SZV; EV; UE3D; SA3D; BRD];
if ~isempty(savepath)
    save(savepath, 'score', '-v7.3');
end

%% Output Draw
% if ~isempty(drawpath)
%     mkdir(drawpath);
%     
%     visualSegmentation(svMap, drawpath);
%     
%     % black & white output
%     mkdir(fullfile(drawpath,'UE'));
%     onMapUE = double(onMapUE);
%     for i=1:size(onMapUE,3)
%         imwrite(onMapUE(:,:,i), fullfile(drawpath,'UE',sprintf('%05d.bw.png',i)));        
%     end
%     mkdir(fullfile(drawpath,'SA'));
%     onMapSA = double(onMapSA);
%     for i=1:size(onMapSA,3)
%         imwrite(onMapSA(:,:,i), fullfile(drawpath,'SA',sprintf('%05d.bw.png',i)));        
%     end
%     
%     % color output    
%     onMapUE = ~logical(onMapUE);
%     for i=1:size(onMapUE,3)
%         colorIM = imread(fullfile(drawpath,sprintf('%05d.png',i)));
%         rIM = colorIM(:,:,1); gIM = colorIM(:,:,2); bIM = colorIM(:,:,3);
%         rIM(onMapUE(:,:,i)) = 0;
%         gIM(onMapUE(:,:,i)) = 0;
%         bIM(onMapUE(:,:,i)) = 0;
%         colorIM = cat(3, rIM, gIM, bIM);
%         imwrite(uint8(colorIM), fullfile(drawpath,'UE',sprintf('%05d.col.png',i)));
%     end
%     onMapSA = ~logical(onMapSA);
%     for i=1:size(onMapSA,3)
%         colorIM = imread(fullfile(drawpath,sprintf('%05d.png',i)));
%         rIM = colorIM(:,:,1); gIM = colorIM(:,:,2); bIM = colorIM(:,:,3);
%         rIM(onMapSA(:,:,i)) = 0;
%         gIM(onMapSA(:,:,i)) = 0;
%         bIM(onMapSA(:,:,i)) = 0;
%         colorIM = cat(3, rIM, gIM, bIM);
%         imwrite(uint8(colorIM), fullfile(drawpath,'SA',sprintf('%05d.col.png',i)));
%     end
%     
%     % Boundary output
%     mkdir(fullfile(drawpath,'BRD'));
%     gtBrWithin = logical(gtBrWithin);
%     for i=1:length(gtPos)
%         colorIM = imread(fullfile(drawpath,sprintf('%05d.png',gtPos(i))));
%         rIM = colorIM(:,:,1); gIM = colorIM(:,:,2); bIM = colorIM(:,:,3);
%         onBR = imdilate(gtBrWithin(:,:,i), strel('disk',1));
%         rIM(onBR) = 0;
%         gIM(onBR) = 0;
%         bIM(onBR) = 0;
%         colorIM = cat(3, rIM, gIM, bIM);
%         imwrite(uint8(colorIM), fullfile(drawpath,'BRD',sprintf('%05d.png',gtPos(i))));
%     end
%     
%     % Single groundtruth output
%     mkdir(fullfile(drawpath,'gt'));
%     gtMap = double(gtMap);
%     for i=1:length(gtPos)
%         imwrite(gtMap(:,:,i), fullfile(drawpath,'gt',sprintf('%05d.png',gtPos(i))));
%     end
% end






