function orgFrames(IN_path, IN_fmt, OUT_path, OUT_fmt, offset, OUT_w, OUT_h, opt)
% This function is used to organize video frames up to 99999 in a folder.
% Things you can do with this function:
% 	1) change frame naming to %05d
% 	2) change frame format
% 	3) resize frame resolusion
%
% AUGUMENTS
% IN_path: input path that contains a set of frames
% IN_fmt: input image format such as 'png'/'ppm'/'jpg'
% OUT_path: output path that saves the output frames
% OUT_fmt: output image format such as 'png'/'ppm'/'jpg'
% offset:
% 	1) 0 -- maintain same frame index as in IN_path
% 	2) 99999 -- output frames in 1:N sequential order
% 	3) others -- OUT_idx = IN_idx + offset
% OUT_w, OUT_h: width and height of output frames
% opt: image resize option 'nearest', 'bicubic' etc.
%
% Author: Chenliang Xu (cxu@nec-labs.com)
% Date: 05/29/2014

% To varify the number of arguments.
if nargin < 4 || nargin == 6 || nargin == 7
	error('Not enough arguments!');
end
fmtFlag = false;
scaleFlag = false;

% To check if the input frames are valid.
imList = dir(fullfile(IN_path, ['*.',IN_fmt]));
imNum = length(imList);
fprintf('Found %d %s frames in folder: %s \n', imNum, IN_fmt, IN_path);
if imNum < 1
	return
end

% To create the output path if not existed.
if ~exist(OUT_path, 'dir')
	mkdir(OUT_path);
end

% To check the fmt
if ~strcmp(IN_fmt, OUT_fmt)
	fmtFlag = true;
end

% To set the offset
if nargin > 4
	if ischar(offset)
		offset = str2double(offset);
	end
else
	offset = 0;
end

% To check image resolusion
if nargin > 5
	if ischar(OUT_w)
		OUT_w = str2double(OUT_w);
	end
	if ischar(OUT_h)
		OUT_h = str2double(OUT_h);
	end
	im = imread(fullfile(IN_path, imList(1).name));
	[IN_h, IN_w, ~] = size(im);
	if IN_h ~= OUT_h || IN_w ~= OUT_w
		scaleFlag = true;
	end
end

% To organize the frames
for i=1:length(imList)
	switch offset
		case 0
			[~, im_name, ~] = fileparts(imList(i).name);
			savepath = fullfile(OUT_path, [im_name,'.',OUT_fmt]);
		case 99999
			savepath = fullfile(OUT_path, sprintf('%05d.%s', i, OUT_fmt));
		otherwise
			[~, im_name, ~] = fileparts(imList(i).name);
			im_idx = str2double(im_name); % to be handle prefix
			im_idx = im_idx + offset;
			savepath = fullfile(OUT_path, sprintf('%05d.%s', im_idx, OUT_fmt));
	end

	if scaleFlag || fmtFlag
		im = imread(fullfile(IN_path, imList(i).name));
		if scaleFlag
			im = imresize(im, [OUT_h, OUT_w], opt);
		end
		imwrite(im, savepath);
	else
		cmd = sprintf('cp %s %s', fullfile(IN_path, imList(i).name), savepath);
		system(cmd);
	end

	if mod(i,100) == 0
		fprintf('processed %d frames ... \n', i);
	end
end

